.pretty_container {
  border-radius: 5px;
  background-color: white;
  margin: 0.5rem;
  padding: 1rem;
  position: relative;
  border: 1px solid #f1f1f1;
  display: flex;
  align-items: center;
  justify-content: center;
}

.indicator_value {
  color: #333333;
}

.indicators {
  display: flex;
  align-items: stretch;
}

.indicator {
  flex: 1;
  padding: 0.75rem 1.75rem;
  text-align: center;
}

.indicator_value {
  font-size: 2rem;
  margin: 1rem 0;
}

.indicator_value p {
  margin: 1rem 0;
}


/* Layouts for different screen sizes */
@media (max-width: 450px) {
  .indicator_value {
    font-size: 1.25rem !important;
  }
  .indicator {
    padding: 0.5rem;
  }
  .button {
    width: 25% !important;
  }
}

@media (max-width: 900px) {
  .indicator {
    align-content: center;
    justify-items: center;
    text-align: center;
  }

  .indicator:first-child {
    margin-right: 0;
  }

  .indicator:last-child {
    margin-left: 0;
  }

  .indicator_value {
    font-size: 1.5rem;
  }
}

@media (min-width: 900px) {
  #lead_grid > .indicators {
    -ms-grid-row: 2;
    -ms-grid-row-span: 1;
    grid-row: 2 / 3;
    -ms-grid-column: 3;
    -ms-grid-column-span: 4;
    grid-column: 3 / 7;
  }

  #opportunity_indicators {
    -ms-grid-column: 3;
    -ms-grid-column-span: 4;
    grid-column: 3 / 7;
    -ms-grid-row: 2;
    -ms-grid-row-span: 1;
    grid-row: 2 / 3;
  }

  #cases_indicators {
    -ms-grid-column: 3;
    -ms-grid-column-span: 4;
    grid-column: 3 / 7;
    -ms-grid-row: 2;
    -ms-grid-row-span: 1;
    grid-row: 2 / 3;
  }
}
